package ru.dbotthepony.kstarbound.defs

import ru.dbotthepony.kommons.io.readUUID
import ru.dbotthepony.kommons.io.writeBinaryString
import ru.dbotthepony.kommons.io.writeUUID
import ru.dbotthepony.kstarbound.io.readInternedString
import ru.dbotthepony.kstarbound.network.syncher.legacyCodec
import ru.dbotthepony.kstarbound.network.syncher.nativeCodec
import ru.dbotthepony.kstarbound.world.UniversePos
import java.io.DataInputStream
import java.io.DataOutputStream
import java.util.UUID

sealed class WorldID {
	abstract fun write(stream: DataOutputStream, isLegacy: Boolean)
	val isLimbo: Boolean get() = this is Limbo

	object Limbo : WorldID() {
		override fun write(stream: DataOutputStream, isLegacy: Boolean) {
			stream.writeByte(0)
		}

		override fun toString(): String {
			return "WorldID.Limbo"
		}
	}

	data class Celestial(val pos: UniversePos) : WorldID() {
		override fun write(stream: DataOutputStream, isLegacy: Boolean) {
			stream.writeByte(1)
			pos.write(stream, isLegacy)
		}

		override fun toString(): String {
			return "WorldID.Celestial[$pos]"
		}
	}

	data class ShipWorld(val uuid: UUID) : WorldID() {
		override fun write(stream: DataOutputStream, isLegacy: Boolean) {
			stream.writeByte(2)
			stream.writeUUID(uuid)
		}

		override fun toString(): String {
			return "WorldID.ShipWorld[${uuid.toString().substring(0, 8)}]"
		}
	}

	data class Instance(val name: String, val uuid: UUID? = null, val threatLevel: Double? = null) : WorldID() {
		override fun write(stream: DataOutputStream, isLegacy: Boolean) {
			stream.writeByte(3)
			stream.writeBinaryString(name)

			stream.writeBoolean(uuid != null)
			if (uuid != null) stream.writeUUID(uuid)

			stream.writeBoolean(threatLevel != null)
			if (threatLevel != null) {
				if (isLegacy) {
					stream.writeFloat(threatLevel.toFloat())
				} else {
					stream.writeDouble(threatLevel)
				}
			}
		}

		override fun toString(): String {
			return "WorldID.Instance[$name, uuid=$uuid, threat level=$threatLevel]"
		}
	}

	companion object {
		val CODEC = nativeCodec(::read, WorldID::write)
		val LEGACY_CODEC = legacyCodec(::read, WorldID::write)

		fun read(stream: DataInputStream, isLegacy: Boolean): WorldID {
			return when (val type = stream.readUnsignedByte()) {
				0 -> Limbo
				1 -> Celestial(UniversePos(stream, isLegacy))
				2 -> ShipWorld(stream.readUUID())
				3 -> {
					val name = stream.readInternedString()
					val uuid = if (stream.readBoolean()) stream.readUUID() else null
					val level: Double?

					if (stream.readBoolean()) {
						if (isLegacy) {
							level = stream.readFloat().toDouble()
						} else {
							level = stream.readDouble()
						}
					} else {
						level = null
					}

					Instance(name, uuid, level)
				}
				else -> throw IllegalArgumentException("Unknown WorldID type $type!")
			}
		}
	}
}
