package ru.dbotthepony.kstarbound.defs.world

import com.google.common.collect.ImmutableList
import com.google.gson.stream.JsonWriter
import ru.dbotthepony.kommons.io.StreamCodec
import ru.dbotthepony.kommons.math.RGBAColor
import ru.dbotthepony.kommons.util.Either
import ru.dbotthepony.kommons.vector.Vector2d
import ru.dbotthepony.kstarbound.defs.AssetPath
import ru.dbotthepony.kstarbound.io.readColor
import ru.dbotthepony.kstarbound.io.writeColor
import ru.dbotthepony.kstarbound.json.builder.IStringSerializable
import ru.dbotthepony.kstarbound.json.builder.JsonFactory
import ru.dbotthepony.kstarbound.util.random.random
import ru.dbotthepony.kstarbound.world.Universe
import ru.dbotthepony.kstarbound.world.UniversePos
import java.io.DataInputStream
import java.io.DataOutputStream

enum class SkyType {
	BARREN,
	ATMOSPHERIC,
	ATMOSPHERELESS,
	ORBITAL,
	WARP,
	SPACE;

	companion object {
		val CODEC = StreamCodec.Enum(SkyType::class.java)
	}
}

enum class FlyingType {
	NONE,
	DISEMBARKING,
	WARP,
	ARRIVING;

	companion object {
		val CODEC = StreamCodec.Enum(FlyingType::class.java)
	}
}

enum class WarpPhase(val stupidassbitch: Int) {
	SLOWING_DOWN(-1),
	MAINTAIN(0),
	SPEEDING_UP(1)
}

enum class SkyOrbiterType(override val jsonName: String) : IStringSerializable {
	SUN("sun"),
	MOON("moon"),
	HORIZON_CLOUD("horizoncloud"),
	SPACE_DEBRIS("scapedebris");
}

@JsonFactory
data class SkyOrbiter(val type: SkyOrbiterType, val scale: Double, val angle: Double, val image: String, val position: Vector2d)

@JsonFactory
data class SkyColoring(
	val mainColor: RGBAColor = RGBAColor.BLACK,

	val morningColors: Pair<RGBAColor, RGBAColor> = RGBAColor.BLACK to RGBAColor.BLACK,
	val dayColors: Pair<RGBAColor, RGBAColor> = RGBAColor.BLACK to RGBAColor.BLACK,
	val eveningColors: Pair<RGBAColor, RGBAColor> = RGBAColor.BLACK to RGBAColor.BLACK,
	val nightColors: Pair<RGBAColor, RGBAColor> = RGBAColor.BLACK to RGBAColor.BLACK,

	val morningLightColor: RGBAColor = RGBAColor.BLACK,
	val dayLightColor: RGBAColor = RGBAColor.BLACK,
	val eveningLightColor: RGBAColor = RGBAColor.BLACK,
	val nightLightColor: RGBAColor = RGBAColor.BLACK,
) {
	fun write(stream: DataOutputStream, isLegacy: Boolean) {
		stream.writeColor(mainColor)

		stream.writeColor(morningColors.first)
		stream.writeColor(morningColors.second)
		stream.writeColor(dayColors.first)
		stream.writeColor(dayColors.second)
		stream.writeColor(eveningColors.first)
		stream.writeColor(eveningColors.second)
		stream.writeColor(nightColors.first)
		stream.writeColor(nightColors.second)

		stream.writeColor(morningLightColor)
		stream.writeColor(dayLightColor)
		stream.writeColor(eveningLightColor)
		stream.writeColor(nightLightColor)
	}

	companion object {
		val BLACK = SkyColoring()

		fun read(stream: DataInputStream, isLegacy: Boolean): SkyColoring {
			val mainColor = stream.readColor()

			val morningColors = stream.readColor() to stream.readColor()
			val dayColors = stream.readColor() to stream.readColor()
			val eveningColors = stream.readColor() to stream.readColor()
			val nightColors = stream.readColor() to stream.readColor()

			val morningLightColor = stream.readColor()
			val dayLightColor = stream.readColor()
			val eveningLightColor = stream.readColor()
			val nightLightColor = stream.readColor()

			return SkyColoring(
				mainColor = mainColor,
				morningColors = morningColors,
				dayColors = dayColors,
				eveningColors = eveningColors,
				nightColors = nightColors,
				morningLightColor = morningLightColor,
				dayLightColor = dayLightColor,
				eveningLightColor = eveningLightColor,
				nightLightColor = nightLightColor,
			)
		}
	}
}

data class SkyWorldHorizon(val center: Vector2d, val scale: Double, val rotation: Double, val layers: List<Pair<String, String>>)

@JsonFactory
data class SkyParameters(
	var skyType: SkyType = SkyType.BARREN,
	var seed: Long = 0L,
	var dayLength: Double? = null,
	var horizonClouds: Boolean = false,
	var skyColoring: Either<SkyColoring, RGBAColor> = Either.left(SkyColoring()),
	var spaceLevel: Double? = null,
	var surfaceLevel: Double? = null,
	var nearbyPlanet: Pair<List<Pair<String, Double>>, Vector2d>? = null,
) {
	companion object {
		suspend fun create(coordinate: UniversePos, universe: Universe): SkyParameters {
			if (coordinate.isSystem)
				throw IllegalArgumentException("$coordinate is system location")

			val params = universe.parameters(coordinate) ?: throw IllegalArgumentException("$universe has nothing at $coordinate!")

			val random = random(params.seed)
			val selfPos = params.coordinate

			val sky = SkyParameters()

			if (selfPos.isSatellite) {
				val planet = universe.parameters(selfPos.parent())

				if (planet != null) {
					val pos = Vector2d(random.nextDouble(), random.nextDouble())
				}
			}

			for (satellitePos in universe.children(selfPos.planet())) {
				if (satellitePos != selfPos) {
					val satellite = universe.parameters(satellitePos)

					if (satellite != null) {
						val pos = Vector2d(random.nextDouble(), random.nextDouble())
					}
				}
			}

			return sky
		}
	}
}

data class SkyGlobalConfig(
	val stars: Stars,
) {
	data class Stars(
		val frames: Int,
		val list: ImmutableList<AssetPath>,
		val hyperlist: ImmutableList<AssetPath>,
	)
}
