package ru.dbotthepony.kstarbound.util.random

import ru.dbotthepony.kstarbound.getValue
import java.util.random.RandomGenerator

// multiply with carry random number generator, as used by original Starbound
// Game's code SHOULD NOT be tied to this generator, random() global function should be used
// What interesting though, is the size of cycle - 256 (8092 bits). Others use much bigger number - 4096
class MWCRandom(seed: Long = System.nanoTime(), cycle: Int = 256, windupIterations: Int = 32) : RandomGenerator {
	private val data = IntArray(cycle)
	private var carry = 0
	private var dataIndex = 0

	var seed: Long = seed
		private set

	init {
		require(windupIterations >= 0) { "Negative amount of windup iterations: $windupIterations" }
		init(seed, windupIterations)
	}

	/**
	 * re-initialize this MWC generator
	 */
	fun init(seed: Long, windupIterations: Int = 0) {
		this.seed = seed
		carry = (seed % MAGIC).toInt()

		data[0] = seed.toInt()
		data[1] = seed.ushr(32).toInt()

		for (i in 2 until data.size) {
			data[i] = 69069 * data[i - 2] + 362437
		}

		dataIndex = data.size - 1

		// initial windup
		for (i in 0 until windupIterations) {
			nextLong()
		}
	}

	fun addEntropy(seed: Long = System.nanoTime()) {
		// Same algo as init, but bitwise xor with existing data
		carry = ((carry.toLong().xor(seed)) % MAGIC).toInt()

		data[0] = data[0].xor(seed.toInt())
		data[1] = data[1].xor(seed.ushr(32).xor(seed).toInt())

		for (i in 2 until data.size) {
			data[i] = data[i].xor(69069 * data[i - 2] + 362437)
		}
	}

	override fun nextLong(): Long {
		dataIndex = (dataIndex + 1) % data.size
		val t = MAGIC.toLong() * data[dataIndex].toLong() + carry.toLong()

		carry = t.ushr(32).toInt()
		data[dataIndex] = t.toInt()

		return t
	}

	companion object {
		const val MAGIC = 809430660

		val GLOBAL: MWCRandom by ThreadLocal.withInitial { MWCRandom() }
	}
}
